import test_ndarray_ext as t
import pytest
import warnings
import importlib
from common import collect

try:
    import numpy as np
    def needs_numpy(x):
        return x
except:
    needs_numpy = pytest.mark.skip(reason="NumPy is required")

try:
    import torch
    def needs_torch(x):
        return x
except:
    needs_torch = pytest.mark.skip(reason="PyTorch is required")

try:
    import tensorflow as tf
    import tensorflow.config
    def needs_tensorflow(x):
        return x
except:
    needs_tensorflow = pytest.mark.skip(reason="TensorFlow is required")

try:
    import jax.numpy as jnp
    def needs_jax(x):
        return x
except:
    needs_jax = pytest.mark.skip(reason="JAX is required")

try:
    import cupy as cp
    def needs_cupy(x):
        return x
except:
    needs_cupy = pytest.mark.skip(reason="CuPy is required")


@needs_numpy
def test01_metadata():
    a = np.zeros(shape=())
    assert t.get_shape(a) == []

    if hasattr(a, '__dlpack__'):
        b = a.__dlpack__()
        assert t.get_shape(b) == []
    else:
        b = None

    with pytest.raises(TypeError) as excinfo:
        # Capsule can only be consumed once
        assert t.get_shape(b) == []
    assert 'incompatible function arguments' in str(excinfo.value)

    a = np.zeros(shape=(3, 4, 5), dtype=np.float64)
    assert t.get_is_valid(a)
    assert t.get_shape(a) == [3, 4, 5]
    assert t.get_size(a) == 60
    assert t.get_nbytes(a) == 60*8
    assert t.get_itemsize(a) == 8
    assert t.check_shape_ptr(a)
    assert t.check_stride_ptr(a)
    if hasattr(a, '__dlpack__'):
        assert t.get_shape(a.__dlpack__()) == [3, 4, 5]
    assert not t.check_float(np.array([1], dtype=np.bool_)) and \
           not t.check_float(np.array([1], dtype=np.uint32)) and \
               t.check_float(np.array([1], dtype=np.float32))
    assert not t.check_bool(np.array([1], dtype=np.uint32)) and \
           not t.check_bool(np.array([1], dtype=np.float32)) and \
               t.check_bool(np.array([1], dtype=np.bool_))

    assert not t.get_is_valid(None)
    assert t.get_size(None) == 0
    assert t.get_nbytes(None) == 0
    assert t.get_itemsize(None) == 0


def test02_docstr():
    assert t.pass_uint32.__doc__ == "pass_uint32(array: ndarray[dtype=uint32]) -> None"
    assert t.get_shape.__doc__ == "get_shape(array: ndarray[writable=False]) -> list"
    assert t.pass_float32.__doc__ == "pass_float32(array: ndarray[dtype=float32]) -> None"
    assert t.pass_complex64.__doc__ == "pass_complex64(array: ndarray[dtype=complex64]) -> None"
    assert t.pass_bool.__doc__ == "pass_bool(array: ndarray[dtype=bool]) -> None"
    assert t.pass_float32_shaped.__doc__ == "pass_float32_shaped(array: ndarray[dtype=float32, shape=(3, *, 4)]) -> None"
    assert t.pass_float32_shaped_ordered.__doc__ == "pass_float32_shaped_ordered(array: ndarray[dtype=float32, shape=(*, *, 4), order='C']) -> None"
    assert t.check_device.__doc__ == ("check_device(arg: ndarray[device='cpu'], /) -> str\n"
                                      "check_device(arg: ndarray[device='cuda'], /) -> str")


@needs_numpy
def test03_constrain_dtype():
    a_u32 = np.array([1], dtype=np.uint32)
    a_f32 = np.array([1], dtype=np.float32)
    a_cf64 = np.array([1+1j], dtype=np.complex64)
    a_bool = np.array([1], dtype=np.bool_)

    t.pass_uint32(a_u32)
    t.pass_float32(a_f32)
    t.pass_complex64(a_cf64)
    t.pass_complex64_const(a_cf64)
    t.pass_bool(a_bool)

    a_f32_const = a_f32.copy()
    a_f32_const.flags.writeable = False
    t.pass_float32_const(a_f32_const)

    a_cf64_const = a_cf64.copy()
    a_cf64_const.flags.writeable = False
    t.pass_complex64_const(a_cf64_const)

    with pytest.raises(TypeError) as excinfo:
        t.pass_uint32(a_f32)
    assert 'incompatible function arguments' in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo:
        t.pass_float32(a_u32)
    assert 'incompatible function arguments' in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo:
        t.pass_complex64(a_u32)
    assert 'incompatible function arguments' in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo:
        t.pass_bool(a_u32)
    assert 'incompatible function arguments' in str(excinfo.value)


@needs_numpy
def test04_constrain_shape():
    t.pass_float32_shaped(np.zeros((3, 0, 4), dtype=np.float32))
    t.pass_float32_shaped(np.zeros((3, 5, 4), dtype=np.float32))

    with pytest.raises(TypeError) as excinfo:
        t.pass_float32_shaped(np.zeros((3, 5), dtype=np.float32))

    with pytest.raises(TypeError) as excinfo:
        t.pass_float32_shaped(np.zeros((2, 5, 4), dtype=np.float32))

    with pytest.raises(TypeError) as excinfo:
        t.pass_float32_shaped(np.zeros((3, 5, 6), dtype=np.float32))

    with pytest.raises(TypeError) as excinfo:
        t.pass_float32_shaped(np.zeros((3, 5, 4, 6), dtype=np.float32))


@needs_numpy
def test05_constrain_order():
    assert t.check_order(np.zeros((3, 5, 4, 6), order='C')) == 'C'
    assert t.check_order(np.zeros((3, 5, 4, 6), order='F')) == 'F'
    assert t.check_order(np.zeros((3, 5, 4, 6), order='C')[:, 2, :, :]) == '?'
    assert t.check_order(np.zeros((3, 5, 4, 6), order='F')[:, 2, :, :]) == '?'


@needs_jax
def test06_constrain_order_jax():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        try:
            c = jnp.zeros((3, 5))
        except:
            pytest.skip('jax is missing')

    z = jnp.zeros((3, 5, 4, 6))
    assert t.check_order(z) == 'C'


@needs_torch
@pytest.mark.filterwarnings
def test07_constrain_order_pytorch():
    try:
        c = torch.zeros(3, 5)
        c.__dlpack__()
    except:
        pytest.skip('pytorch is missing')

    f = c.t().contiguous().t()
    assert t.check_order(c) == 'C'
    assert t.check_order(f) == 'F'
    assert t.check_order(c[:, 2:5]) == '?'
    assert t.check_order(f[1:3, :]) == '?'
    assert t.check_device(torch.zeros(3, 5)) == 'cpu'
    if torch.cuda.is_available():
        assert t.check_device(torch.zeros(3, 5, device='cuda')) == 'cuda'


@needs_tensorflow
def test08_constrain_order_tensorflow():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        try:
            c = tf.zeros((3, 5))
        except:
            pytest.skip('tensorflow is missing')

    assert t.check_order(c) == 'C'


@needs_numpy
def test09_write_from_cpp():
    x = np.zeros(10, dtype=np.float32)
    t.initialize(x)
    assert np.all(x == np.arange(10, dtype=np.float32))

    x = np.zeros((10, 3), dtype=np.float32)
    t.initialize(x)
    assert np.all(x == np.arange(30, dtype=np.float32).reshape(10, 3))


@needs_numpy
def test10_implicit_conversion():
    t.implicit(np.zeros((2, 2), dtype=np.uint32))
    t.implicit(np.zeros((2, 2, 10), dtype=np.float32)[:, :, 4])
    t.implicit(np.zeros((2, 2, 10), dtype=np.uint32)[:, :, 4])
    t.implicit(np.zeros((2, 2, 10), dtype=np.bool_)[:, :, 4])

    with pytest.raises(TypeError) as excinfo:
        t.noimplicit(np.zeros((2, 2), dtype=np.bool_))

    with pytest.raises(TypeError) as excinfo:
        t.noimplicit(np.zeros((2, 2), dtype=np.uint32))

    with pytest.raises(TypeError) as excinfo:
        t.noimplicit(np.zeros((2, 2, 10), dtype=np.float32)[:, :, 4])


@needs_torch
def test11_implicit_conversion_pytorch():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        try:
            c = torch.zeros(3, 5)
            c.__dlpack__()
        except:
            pytest.skip('pytorch is missing')

    t.implicit(torch.zeros(2, 2, dtype=torch.int32))
    t.implicit(torch.zeros(2, 2, 10, dtype=torch.float32)[:, :, 4])
    t.implicit(torch.zeros(2, 2, 10, dtype=torch.int32)[:, :, 4])

    with pytest.raises(TypeError) as excinfo:
        t.noimplicit(torch.zeros(2, 2, dtype=torch.int32))

    with pytest.raises(TypeError) as excinfo:
        t.noimplicit(torch.zeros(2, 2, 10, dtype=torch.float32)[:, :, 4])


@needs_tensorflow
def test12_implicit_conversion_tensorflow():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        try:
            c = tf.zeros((3, 5))
        except:
            pytest.skip('tensorflow is missing')

        t.implicit(tf.zeros((2, 2), dtype=tf.int32))
        t.implicit(tf.zeros((2, 2, 10), dtype=tf.float32)[:, :, 4])
        t.implicit(tf.zeros((2, 2, 10), dtype=tf.int32)[:, :, 4])
        t.implicit(tf.zeros((2, 2, 10), dtype=tf.bool)[:, :, 4])

        with pytest.raises(TypeError) as excinfo:
            t.noimplicit(tf.zeros((2, 2), dtype=tf.int32))

        with pytest.raises(TypeError) as excinfo:
            t.noimplicit(tf.zeros((2, 2), dtype=tf.bool))


@needs_jax
def test13_implicit_conversion_jax():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        try:
            c = jnp.zeros((3, 5))
        except:
            pytest.skip('jax is missing')

    t.implicit(jnp.zeros((2, 2), dtype=jnp.int32))
    t.implicit(jnp.zeros((2, 2, 10), dtype=jnp.float32)[:, :, 4])
    t.implicit(jnp.zeros((2, 2, 10), dtype=jnp.int32)[:, :, 4])
    t.implicit(jnp.zeros((2, 2, 10), dtype=jnp.bool_)[:, :, 4])

    with pytest.raises(TypeError) as excinfo:
        t.noimplicit(jnp.zeros((2, 2), dtype=jnp.int32))

    with pytest.raises(TypeError) as excinfo:
        t.noimplicit(jnp.zeros((2, 2), dtype=jnp.uint8))


def test14_destroy_capsule():
    collect()
    dc = t.destruct_count()
    a = t.return_dlpack()
    assert dc == t.destruct_count()
    del a
    collect()
    assert t.destruct_count() - dc == 1


@needs_numpy
def test15_consume_numpy():
    collect()
    class wrapper:
        def __init__(self, value):
            self.value = value
        def __dlpack__(self):
            return self.value
    dc = t.destruct_count()
    a = t.return_dlpack()
    if hasattr(np, '_from_dlpack'):
        x = np._from_dlpack(wrapper(a))
    elif hasattr(np, 'from_dlpack'):
        x = np.from_dlpack(wrapper(a))
    else:
        pytest.skip('your version of numpy is too old')

    del a
    collect()
    assert x.shape == (2, 4)
    assert np.all(x == [[1, 2, 3, 4], [5, 6, 7, 8]])
    assert dc == t.destruct_count()
    del x
    collect()
    assert t.destruct_count() - dc == 1


@needs_numpy
def test16_passthrough():
    a = t.ret_numpy()
    b = t.passthrough(a)
    assert a is b

    a = np.array([1,2,3])
    b = t.passthrough(a)
    assert a is b

    a = None
    with pytest.raises(TypeError) as excinfo:
        b = t.passthrough(a)
    assert 'incompatible function arguments' in str(excinfo.value)
    b = t.passthrough_arg_none(a)
    assert a is b


@needs_numpy
def test17_return_numpy():
    collect()
    dc = t.destruct_count()
    x = t.ret_numpy()
    assert x.shape == (2, 4)
    assert np.all(x == [[1, 2, 3, 4], [5, 6, 7, 8]])
    del x
    collect()
    assert t.destruct_count() - dc == 1


@needs_torch
def test18_return_pytorch():
    try:
        c = torch.zeros(3, 5)
    except:
        pytest.skip('pytorch is missing')
    collect()
    dc = t.destruct_count()
    x = t.ret_pytorch()
    assert x.shape == (2, 4)
    assert torch.all(x == torch.tensor([[1, 2, 3, 4], [5, 6, 7, 8]]))
    del x
    collect()
    assert t.destruct_count() - dc == 1


@needs_jax
def test19_return_jax():
    collect()
    dc = t.destruct_count()
    x = t.ret_jax()
    assert x.shape == (2, 4)
    assert jnp.all(x == jnp.array([[1,2,3,4], [5,6,7,8]], dtype=jnp.float32))
    del x
    collect()
    assert t.destruct_count() - dc == 1


@needs_tensorflow
def test20_return_tensorflow():
    collect()
    dc = t.destruct_count()
    x = t.ret_tensorflow()
    assert x.get_shape().as_list() == [2, 4]
    assert tf.math.reduce_all(
               x == tf.constant([[1,2,3,4], [5,6,7,8]], dtype=tf.float32))
    del x
    collect()
    assert t.destruct_count() - dc == 1


@needs_numpy
def test21_return_array_scalar():
    collect()
    dc = t.destruct_count()
    x = t.ret_array_scalar()
    assert np.array_equal(x, np.array(1))
    del x
    collect()
    assert t.destruct_count() - dc == 1


# See PR #162
@needs_torch
def test22_single_and_empty_dimension_pytorch():
    a = torch.ones((1,100,1025), dtype=torch.float32)
    t.noop_3d_c_contig(a)
    a = torch.ones((100,1,1025), dtype=torch.float32)
    t.noop_3d_c_contig(a)
    a = torch.ones((0,100,1025), dtype=torch.float32)
    t.noop_3d_c_contig(a)
    a = torch.ones((100,0,1025), dtype=torch.float32)
    t.noop_3d_c_contig(a)
    a = torch.ones((100,1025,0), dtype=torch.float32)
    t.noop_3d_c_contig(a)
    a = torch.ones((100,0,0), dtype=torch.float32)
    t.noop_3d_c_contig(a)
    a = torch.ones((0,0,0), dtype=torch.float32)
    t.noop_3d_c_contig(a)


# See PR #162
@needs_numpy
def test23_single_and_empty_dimension_numpy():
    a = np.ones((1,100,1025), dtype=np.float32)
    t.noop_3d_c_contig(a)
    a = np.ones((100,1,1025), dtype=np.float32)
    t.noop_3d_c_contig(a)
    a = np.ones((0,100,1025), dtype=np.float32)
    t.noop_3d_c_contig(a)
    a = np.ones((100,0,1025), dtype=np.float32)
    t.noop_3d_c_contig(a)
    a = np.ones((100,1025,0), dtype=np.float32)
    t.noop_3d_c_contig(a)
    a = np.ones((100,0,0), dtype=np.float32)
    t.noop_3d_c_contig(a)
    a = np.ones((0,0,0), dtype=np.float32)
    t.noop_3d_c_contig(a)


# See PR #162
@needs_torch
def test24_single_and_empty_dimension_fortran_order_pytorch():
    # This idiom creates a pytorch 2D tensor in column major (aka, 'F') ordering
    a = torch.ones((0,100), dtype=torch.float32).t().contiguous().t()
    t.noop_2d_f_contig(a)
    a = torch.ones((100,0), dtype=torch.float32).t().contiguous().t()
    t.noop_2d_f_contig(a)
    a = torch.ones((1,100), dtype=torch.float32).t().contiguous().t()
    t.noop_2d_f_contig(a)
    a = torch.ones((100,1), dtype=torch.float32).t().contiguous().t()
    t.noop_2d_f_contig(a)


@needs_numpy
def test25_ro_array():
    a = np.array([1, 2], dtype=np.float32)
    assert t.accept_ro(a) == 1
    assert t.accept_rw(a) == 1
    a.setflags(write=False)
    assert t.accept_ro(a) == 1
    with pytest.raises(TypeError) as excinfo:
        t.accept_rw(a)
    assert 'incompatible function arguments' in str(excinfo.value)


@needs_numpy
def test26_return_ro():
    x = t.ret_numpy_const_ref()
    y = t.ret_numpy_const_ref_f()
    assert t.ret_numpy_const_ref.__doc__  == 'ret_numpy_const_ref() -> numpy.ndarray[dtype=float32, shape=(2, 4), order=\'C\', writable=False]'
    assert t.ret_numpy_const_ref_f.__doc__  == 'ret_numpy_const_ref_f() -> numpy.ndarray[dtype=float32, shape=(2, 4), order=\'F\', writable=False]'
    assert x.shape == (2, 4)
    assert y.shape == (2, 4)
    assert np.all(x == [[1, 2, 3, 4], [5, 6, 7, 8]])
    assert np.all(y == [[1, 3, 5, 7], [2, 4, 6, 8]])
    with pytest.raises(ValueError) as excinfo:
        x[0,0] =1
    assert 'read-only' in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        y[0,0] =1
    assert 'read-only' in str(excinfo.value)


@needs_numpy
def test27_check_numpy():
    assert t.check(np.zeros(1))


@needs_torch
def test28_check_torch():
    assert t.check(torch.zeros((1)))


@needs_tensorflow
def test29_check_tensorflow():
    assert t.check(tf.zeros((1)))


@needs_jax
def test30_check_jax():
    assert t.check(jnp.zeros((1)))


@needs_numpy
def test31_rv_policy():
    def p(a):
        return a.__array_interface__['data']

    x1 = t.ret_numpy_const_ref()
    x2 = t.ret_numpy_const_ref()
    y1 = t.ret_numpy_const()
    y2 = t.ret_numpy_const()

    z1 = t.passthrough(y1)
    z2 = t.passthrough(y2)
    q1 = t.passthrough_copy(y1)
    q2 = t.passthrough_copy(y2)

    assert p(x1) == p(x2)
    assert p(y1) != p(y2)

    assert z1 is y1
    assert z2 is y2
    assert q1 is not y1
    assert q2 is not y2
    assert p(q1) != p(y1)
    assert p(q2) != p(y2)


@needs_numpy
def test32_reference_internal():
    collect()
    dc = t.destruct_count()
    c = t.Cls()

    v1_a = c.f1()
    v1_b = c.f1()
    v2_a = c.f2()
    v2_b = c.f2()
    del c

    assert np.all(v1_a == np.arange(10, dtype=np.float32))
    assert np.all(v1_b == np.arange(10, dtype=np.float32))

    v1_a += 1
    v1_b += 2

    assert np.all(v1_a == np.arange(10, dtype=np.float32) + 1)
    assert np.all(v1_b == np.arange(10, dtype=np.float32) + 2)
    del v1_a
    del v1_b

    assert np.all(v2_a == np.arange(10, dtype=np.float32))
    assert np.all(v2_b == np.arange(10, dtype=np.float32))

    v2_a += 1
    v2_b += 2

    assert np.all(v2_a == np.arange(10, dtype=np.float32) + 3)
    assert np.all(v2_b == np.arange(10, dtype=np.float32) + 3)

    del v2_a
    collect()
    assert t.destruct_count() == dc

    del v2_b
    collect()
    dc += 1
    assert t.destruct_count() == dc

    for i in range(2):
        c2 = t.Cls()

        if i == 0:
            v3_a = c2.f1_ri()
            v3_b = c2.f1_ri()
        else:
            v3_a = c2.f2_ri()
            v3_b = c2.f2_ri()
        del c2

        assert np.all(v3_a == np.arange(10, dtype=np.float32))
        assert np.all(v3_b == np.arange(10, dtype=np.float32))

        v3_a += 1
        v3_b += 2

        assert np.all(v3_a == np.arange(10, dtype=np.float32) + 3)
        assert np.all(v3_b == np.arange(10, dtype=np.float32) + 3)
        del v3_a

        collect()
        assert t.destruct_count() == dc

        del v3_b
        collect()
        dc += 1
        assert t.destruct_count() == dc

    c3 = t.Cls()
    c3_t = (c3,)
    with pytest.raises(RuntimeError) as excinfo:
        c3.f3_ri(c3_t)

    msg = 'nanobind::detail::ndarray_export(): reference_internal policy cannot be applied (ndarray already has an owner)'
    assert msg in str(excinfo.value)


@needs_numpy
def test33_force_contig_numpy():
    a = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    b = t.make_contig(a)
    assert b is a
    a = a.T
    b = t.make_contig(a)
    assert b is not a
    assert np.all(b == a)

@needs_torch
@pytest.mark.filterwarnings
def test34_force_contig_pytorch():
    a = torch.tensor([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    b = t.make_contig(a)
    assert b is a
    a = a.T
    b = t.make_contig(a)
    assert b is not a
    assert torch.all(b == a)


@needs_numpy
def test35_view():
    # 1
    x1 = np.array([[1,2],[3,4]], dtype=np.float32)
    x2 = np.array([[1,2],[3,4]], dtype=np.float64)
    assert np.allclose(x1, x2)
    t.fill_view_1(x1)
    assert np.allclose(x1, x2*2)
    t.fill_view_1(x2)
    assert np.allclose(x1, x2*2)

    # 2
    x1 = np.zeros((3, 4), dtype=np.float32, order='C')
    x2 = np.zeros((3, 4), dtype=np.float32, order='F')
    t.fill_view_2(x1)
    t.fill_view_2(x2)
    x3 = np.zeros((3, 4), dtype=np.float32, order='C')
    t.fill_view_3(x3)
    x4 = np.zeros((3, 4), dtype=np.float32, order='F')
    t.fill_view_4(x4)

    assert np.all(x1 == x2) and np.all(x2 == x3) and np.all(x3 == x4)

    # 3
    x1 = np.array([[1+2j, 3+4j], [5+6j, 7+8j]], dtype=np.complex64)
    x2 = x1 * 2
    t.fill_view_1(x1.view(np.float32))
    assert np.allclose(x1, x2)
    x2 = x1 * (-1+2j)
    t.fill_view_5(x1)
    assert np.allclose(x1, x2)
    x2 = -x2;
    t.fill_view_6(x1)
    assert np.allclose(x1, x2)


@needs_numpy
def test36_half():
    if not hasattr(t, 'ret_numpy_half'):
        pytest.skip('half precision test is missing')
    x = t.ret_numpy_half()
    assert x.dtype == np.float16
    assert x.shape == (2, 4)
    assert np.all(x == [[1, 2, 3, 4], [5, 6, 7, 8]])

@needs_numpy
def test37_cast():
    a = t.cast(False)
    b = t.cast(True)
    assert a.ndim == 0 and b.ndim == 0
    assert a.dtype == np.int32 and b.dtype == np.float32
    assert a == 1 and b == 1


@needs_numpy
def test38_complex_decompose():
    x1 = np.array([1 + 2j, 3 + 4j, 5 + 6j], dtype=np.complex64)

    assert np.all(x1.real == np.array([1, 3, 5], dtype=np.float32))
    assert np.all(x1.imag == np.array([2, 4, 6], dtype=np.float32))


@needs_numpy
@pytest.mark.parametrize("variant", [1, 2])
def test_uint32_complex_do_not_convert(variant):
    if variant == 1:
        arg = 1
    else:
        arg = np.uint32(1)
    data = np.array([1.0 + 2.0j, 3.0 + 4.0j])
    t.set_item(data, arg)
    data2 = np.array([123, 3.0 + 4.0j])
    assert np.all(data == data2)


@needs_numpy
def test40_check_generic():
    class DLPackWrapper:
        def __init__(self, o):
            self.o = o
        def __dlpack__(self):
            return self.o.__dlpack__()

    arr = DLPackWrapper(np.zeros((1)))
    assert t.check(arr)


@needs_numpy
def test41_noninteger_stride():
    a = np.array([[1, 2, 3, 4, 0, 0], [5, 6, 7, 8, 0, 0]], dtype=np.float32)
    s = a[:, 0:4]  # slice
    t.pass_float32(s)
    assert t.get_stride(s, 0) == 6;
    assert t.get_stride(s, 1) == 1;
    try:
        v = s.view(np.complex64)
    except:
        pytest.skip('your version of numpy is too old')
    t.pass_complex64(v)
    assert t.get_stride(v, 0) == 3;
    assert t.get_stride(v, 1) == 1;

    a = np.array([[1, 2, 3, 4, 0], [5, 6, 7, 8, 0]], dtype=np.float32)
    s = a[:, 0:4]  # slice
    t.pass_float32(s)
    assert t.get_stride(s, 0) == 5;
    assert t.get_stride(s, 1) == 1;
    v = s.view(np.complex64)
    with pytest.raises(TypeError) as excinfo:
        t.pass_complex64(v)
    assert 'incompatible function arguments' in str(excinfo.value)
    with pytest.raises(TypeError) as excinfo:
        t.get_stride(v, 0);
    assert 'incompatible function arguments' in str(excinfo.value)


@needs_numpy
def test42_const_qualifiers_numpy():
    a = np.array([0, 0, 0, 3.14159, 0], dtype=np.float64)
    assert t.check_rw_by_value(a);
    assert a[1] == 1.414214;
    assert t.check_rw_by_value_float64(a);
    assert a[2] == 2.718282;
    assert a[4] == 16.0;
    assert t.check_ro_by_value_ro(a);
    assert t.check_ro_by_value_const_float64(a);
    a.setflags(write=False)
    assert t.check_ro_by_value_ro(a);
    assert t.check_ro_by_value_const_float64(a);
    assert a[0] == 0.0;
    assert a[3] == 3.14159;

    a = np.array([0, 0, 0, 3.14159, 0], dtype=np.float64)
    assert t.check_rw_by_const_ref(a);
    assert a[1] == 1.414214;
    assert t.check_rw_by_const_ref_float64(a);
    assert a[2] == 2.718282;
    assert a[4] == 16.0;
    assert t.check_ro_by_const_ref_ro(a);
    assert t.check_ro_by_const_ref_const_float64(a);
    a.setflags(write=False)
    assert t.check_ro_by_const_ref_ro(a);
    assert t.check_ro_by_const_ref_const_float64(a);
    assert a[0] == 0.0;
    assert a[3] == 3.14159;

    a = np.array([0, 0, 0, 3.14159, 0], dtype=np.float64)
    assert t.check_rw_by_rvalue_ref(a);
    assert a[1] == 1.414214;
    assert t.check_rw_by_rvalue_ref_float64(a);
    assert a[2] == 2.718282;
    assert a[4] == 16.0;
    assert t.check_ro_by_rvalue_ref_ro(a);
    assert t.check_ro_by_rvalue_ref_const_float64(a);
    a.setflags(write=False)
    assert t.check_ro_by_rvalue_ref_ro(a);
    assert t.check_ro_by_rvalue_ref_const_float64(a);
    assert a[0] == 0.0;
    assert a[3] == 3.14159;


@needs_torch
def test43_const_qualifiers_pytorch():
    a = torch.tensor([0, 0, 0, 3.14159, 0], dtype=torch.float64)
    assert t.check_rw_by_value(a);
    assert a[1] == 1.414214;
    assert t.check_rw_by_value_float64(a);
    assert a[2] == 2.718282;
    assert a[4] == 16.0;
    assert t.check_ro_by_value_ro(a);
    assert t.check_ro_by_value_const_float64(a);
    assert a[0] == 0.0;
    assert a[3] == 3.14159;

    a = torch.tensor([0, 0, 0, 3.14159, 0], dtype=torch.float64)
    assert t.check_rw_by_const_ref(a);
    assert a[1] == 1.414214;
    assert t.check_rw_by_const_ref_float64(a);
    assert a[2] == 2.718282;
    assert a[4] == 16.0;
    assert t.check_ro_by_const_ref_ro(a);
    assert t.check_ro_by_const_ref_const_float64(a);
    assert a[0] == 0.0;
    assert a[3] == 3.14159;

    a = torch.tensor([0, 0, 0, 3.14159, 0], dtype=torch.float64)
    assert t.check_rw_by_rvalue_ref(a);
    assert a[1] == 1.414214;
    assert t.check_rw_by_rvalue_ref_float64(a);
    assert a[2] == 2.718282;
    assert a[4] == 16.0;
    assert t.check_ro_by_rvalue_ref_ro(a);
    assert t.check_ro_by_rvalue_ref_const_float64(a);
    assert a[0] == 0.0;
    assert a[3] == 3.14159;


@needs_cupy
@pytest.mark.filterwarnings
def test44_constrain_order_cupy():
    try:
        c = cp.zeros((3, 5))
        c.__dlpack__()
    except:
        pytest.skip('cupy is missing')

    f = cp.asarray(c, order="F")
    assert t.check_order(c) == 'C'
    assert t.check_order(f) == 'F'
    assert t.check_order(c[:, 2:5]) == '?'
    assert t.check_order(f[1:3, :]) == '?'
    assert t.check_device(cp.zeros((3, 5))) == 'cuda'


@needs_cupy
def test45_implicit_conversion_cupy():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        try:
            c = cp.zeros((3, 5))
        except:
            pytest.skip('cupy is missing')

    t.implicit(cp.zeros((2, 2), dtype=cp.int32))
    t.implicit(cp.zeros((2, 2, 10), dtype=cp.float32)[:, :, 4])
    t.implicit(cp.zeros((2, 2, 10), dtype=cp.int32)[:, :, 4])
    t.implicit(cp.zeros((2, 2, 10), dtype=cp.bool_)[:, :, 4])

    with pytest.raises(TypeError) as excinfo:
        t.noimplicit(cp.zeros((2, 2), dtype=cp.int32))

    with pytest.raises(TypeError) as excinfo:
        t.noimplicit(cp.zeros((2, 2), dtype=cp.uint8))


@needs_numpy
def test46_implicit_conversion_contiguous_complex():
    # Test fix for issue #709
    import numpy as np

    c_f32 = np.random.rand(10, 10)
    c_c64 = c_f32.astype(np.complex64)

    assert c_f32.flags['C_CONTIGUOUS']
    assert c_c64.flags['C_CONTIGUOUS']

    def test_conv(x):
        y = t.test_implicit_conversion(x)
        assert np.all(x == y)
        assert y.flags['C_CONTIGUOUS']

    test_conv(c_f32)
    test_conv(c_c64)

    nc_f32 = c_f32.T
    nc_c64 = c_c64.T

    assert not nc_f32.flags['C_CONTIGUOUS']
    assert not nc_c64.flags['C_CONTIGUOUS']

    test_conv(nc_f32)
    test_conv(nc_c64)


@needs_numpy
def test_47_ret_infer():
    import numpy as np
    assert np.all(t.ret_infer_c() == [[1, 2, 3, 4], [5, 6, 7, 8]])
    assert np.all(t.ret_infer_f() == [[1, 3, 5, 7], [2, 4, 6, 8]])


@needs_numpy
def test48_test_matrix4f():
    a = t.Matrix4f()
    ad = a.data()
    bd = a.data()
    for i in range(16):
        ad[i%4, i//4] = i
    del a, ad
    for i in range(16):
        assert bd[i%4, i//4] == i


@needs_numpy
def test49_test_matrix4f_ref():
    assert t.Matrix4f.data_ref.__doc__.replace('data_ref', 'data') == t.Matrix4f.data.__doc__

    a = t.Matrix4f()
    ad = a.data_ref()
    bd = a.data_ref()
    for i in range(16):
        ad[i%4, i//4] = i
    del a, ad
    for i in range(16):
        assert bd[i%4, i//4] == i


@needs_numpy
def test50_test_matrix4f_copy():
    assert t.Matrix4f.data_ref.__doc__.replace('data_ref', 'data') == t.Matrix4f.data.__doc__

    a = t.Matrix4f()
    ad = a.data_ref()
    for i in range(16):
        ad[i%4, i//4] = i
    bd = a.data_copy()
    for i in range(16):
        ad[i%4, i//4] = 0
    del a, ad
    for i in range(16):
        assert bd[i%4, i//4] == i


@needs_numpy
def test51_return_from_stack():
    import numpy as np
    assert np.all(t.ret_from_stack_1() == [1,2,3])
    assert np.all(t.ret_from_stack_2() == [1,2,3])

@needs_numpy
def test52_accept_np_both_true_contig():
    import numpy as np
    a = np.zeros((2, 1), dtype=np.float32)
    assert a.flags['C_CONTIGUOUS'] and a.flags['F_CONTIGUOUS']
    t.accept_np_both_true_contig_a(a)
    t.accept_np_both_true_contig_c(a)
    t.accept_np_both_true_contig_f(a)


@needs_numpy
def test53_issue_930():
    import numpy as np
    wrapper = t.Wrapper(np.ones(3, dtype=np.float32))
    assert wrapper.value[0] == 1



from collections.abc import Callable
from typing import TypeVar, overload


class AClass:
    STATIC_VAR: int = 5

    class NestedClass:
        pass

    def __init__(self, x): ...

    def method(self, x: str): ...

    @staticmethod
    def static_method(x): ...

    @classmethod
    def class_method(cls, x): ...

    @overload
    def overloaded(self, x: int) -> None:
        """docstr 1"""

    @overload
    def overloaded(self, x: str) -> None:
        """docstr 2"""

    @overload
    def overloaded_2(self, x: int) -> None:
        """docstr 3"""

    @overload
    def overloaded_2(self, x: str) -> None: ...

C: int = 123

T = TypeVar("T")

def f1(a, b, c, /):
    """docstring"""

def f2(a=(3, (4, 5)), /, b=(1, 2), *, c=4):
    """docstring 2"""

def f3(*args, **kwargs): ...

def f4() -> Callable[[T], T]: ...



#!/usr/bin/env python3

# Builds demumble for Mac, Linux, Windows.  Must run on a Mac.
# Needs a chromium checkout at ~/src/chrome/src that was synced with
# target_os=['win'] to get the Windows toolchain, and to get lld.
# You must run `build/linux/sysroot_scripts/install-sysroot.py --arch amd64`
# once to get the linux toolchain.

# Also needs a GN build of llvm at ~/src/llvm-project/out/gn for llvm-strip
# for stripping the Linux binary.

# Run this in the demumble root directory while on the "release" branch.
# It'll create subcirectories "buildlinux", "buildmac", "buildwin" to build
# for each platforms, and then puts the final built products in
# demumble-{linux,mac,win}.zip.

# Runs demumble_test.py on mac at the end, but best make sure it passes on
# on all 3 platforms before running this script.

# https://gitlab.kitware.com/cmake/community/wikis/doc/cmake/CrossCompiling has
# some documentation on cross builds with cmake.

import contextlib
import json
import glob
import os
import subprocess
import sys

crsrc = os.path.join(os.path.expanduser('~'), 'src/chrome/src')
if len(sys.argv) > 1:
  crsrc = os.path.abspath(sys.argv[1])
clangcl = crsrc + '/third_party/llvm-build/Release+Asserts/bin/clang-cl'
clangxx = crsrc + '/third_party/llvm-build/Release+Asserts/bin/clang++'
lldlink = crsrc + '/third_party/llvm-build/Release+Asserts/bin/lld-link'

# FIXME: https://chromium-review.googlesource.com/c/chromium/src/+/1214943
# has a way to build eu-strip on macOS, which is arguably a smaller dep
# than llvm-strip.
llvm_strip = os.path.join(os.path.expanduser('~'),
                          'src/llvm-project/out/gn/bin/llvm-strip')

platform = 'mac' if sys.platform == 'darwin' else 'linux'
cmake = ('/Applications/CMake.app/Contents/bin/cmake' if platform == 'mac' else
         '/usr/bin/cmake')
call_cmake = [cmake, '-GNinja', '..', '-DCMAKE_BUILD_TYPE=Release']


@contextlib.contextmanager
def buildir(newdir):
    """Creates newdir if it doesn't exist yet and temporarily sets cwd to it."""
    newdir = os.path.join(os.path.dirname(__file__), newdir)
    if not os.path.isdir(newdir):
        os.mkdir(newdir)  # Intentionally not deleted.
    prevdir = os.getcwd()
    os.chdir(newdir)
    try:
        yield
    finally:
        os.chdir(prevdir)

subprocess.check_call(['rm', '-rf', 'buildlinux', 'buildmac', 'buildwin'])
subprocess.check_call(
    ['rm', '-f', 'demumble-linux.zip', 'demumble-mac.zip', 'demumble-win.zip'])
devnull = open(os.devnull, 'w')

# Linux.
linux_sysroot = crsrc + '/build/linux/debian_sid_amd64-sysroot'
cflags = [ '--target=x86_64-linux-gnu' ]
ldflags = ['-fuse-ld=lld'] + cflags
with buildir('buildlinux'):
    print('building linux')
    subprocess.check_call(call_cmake + [
        '-DCMAKE_CXX_COMPILER=' + clangxx,
        '-DCMAKE_CXX_FLAGS=' + ' '.join(cflags),
        '-DCMAKE_EXE_LINKER_FLAGS=' + ' '.join(ldflags),
        '-DCMAKE_SYSROOT=' + linux_sysroot,
        '-DCMAKE_SYSTEM_NAME=Linux',
        ], stdout=devnull)
    subprocess.check_call(['ninja', 'demumble'])
    subprocess.check_call([llvm_strip, 'demumble'])
    subprocess.check_call(['zip', '-q9', 'demumble-linux.zip', 'demumble'])
    subprocess.check_call(['mv', 'demumble-linux.zip', '..'])

# Mac.
mac_sysroot = (crsrc + '/build/mac_files/xcode_binaries/Contents/Developer' +
                       '/Platforms/MacOSX.platform/Developer/SDKs/MacOSX.sdk')
if platform == 'mac' and not os.path.isdir(mac_sysroot):
    mac_sysroot_flag = []
else:
    mac_sysroot_flag = [ '-DCMAKE_OSX_SYSROOT=' + mac_sysroot ]
cflags = [ '--target=apple-macos', '-mmacosx-version-min=10.9' ]
ldflags = ['-fuse-ld=lld'] + cflags
with buildir('buildmac'):
    print('building mac')
    subprocess.check_call(call_cmake + mac_sysroot_flag + [
        '-DCMAKE_CXX_COMPILER=' + clangxx,
        '-DCMAKE_CXX_FLAGS=' + ' '.join(cflags),
        '-DCMAKE_EXE_LINKER_FLAGS=' + ' '.join(ldflags),
        '-DCMAKE_OSX_ARCHITECTURES=arm64;x86_64',
        '-DCMAKE_SYSTEM_NAME=Darwin',
        ], stdout=devnull)
    subprocess.check_call(['ninja', 'demumble'])
    subprocess.check_call([llvm_strip, 'demumble'])
    subprocess.check_call(['zip', '-q9', 'demumble-mac.zip', 'demumble'])
    subprocess.check_call(['mv', 'demumble-mac.zip', '..'])

# Win.
win_sysroot = glob.glob(
    crsrc + '/third_party/depot_tools/win_toolchain/vs_files/*')[0]
cflags = ['--target=x86_64-pc-windows', '/winsysroot' + win_sysroot]
# Without /manifest:no, cmake creates a default manifest file -- and
# explicitly calls mt.exe (which we don't have in a cross build).
# This also removes a dependency on rc.exe -- without this we'd also
# have to set CMAKE_TRY_COMPILE_TARGET_TYPE=STATIC_LIBRARY.
# TODO: Remove /machine:x64 once crbug.com/1300005 is fixed.
ldflags = ['/manifest:no', '/winsysroot:' + win_sysroot, '/machine:x64']
with buildir('buildwin'):
    print('building windows')
    subprocess.check_call(call_cmake + [
        '-DCMAKE_CXX_COMPILER=' + clangcl,
        '-DCMAKE_CXX_FLAGS=' + ' '.join(cflags),
        '-DCMAKE_EXE_LINKER_FLAGS=' + ' '.join(ldflags),
        '-DCMAKE_LINKER=' + lldlink,
        '-DCMAKE_SYSTEM_NAME=Windows',
        ], stdout=devnull)
    subprocess.check_call(['ninja', 'demumble'])
    # No stripping on Windows.
    subprocess.check_call(['zip', '-q9', 'demumble-win.zip', 'demumble.exe'])
    subprocess.check_call(['mv', 'demumble-win.zip', '..'])

# Copy over linux or mac binary and run tests.
print(f'running tests (on {platform})')
# https://developer.apple.com/documentation/security/updating_mac_software
subprocess.check_call(f'rm -f demumble && cp build{platform}/demumble .',
                      shell=True)
subprocess.check_call(['./demumble_test.py'])

# Show zip files.
subprocess.check_call('ls -hl *.zip', shell=True)
subprocess.check_call(['./demumble', '--version'])



import sys
import test_holders_ext as t
import pytest
from common import collect


@pytest.fixture
def clean():
    collect()
    t.reset()

# ------------------------------------------------------------------

def test01_create(clean):
    e = t.Example(123)
    assert e.value == 123
    assert t.query_shared_1(e) == 123
    assert t.query_shared_2(e) == 123
    assert t.query_shared_3(e) == 123
    del e
    collect()
    assert t.stats() == (1, 1)


def test02_sharedptr_from_python(clean):
    e = t.Example(234)
    w = t.SharedWrapper(e)
    assert w.ptr is e
    del e
    collect()
    assert t.stats() == (1, 0)
    del w
    collect()
    assert t.stats() == (1, 1)

    w = t.SharedWrapper(t.Example(234))
    collect()
    assert t.stats() == (2, 1)
    w.ptr = t.Example(0)
    collect()
    assert t.stats() == (3, 2)
    del w
    collect()
    assert t.stats() == (3, 3)


def test03_sharedptr_from_cpp(clean):
    e = t.Example.make(5)
    assert t.passthrough(e) is e
    assert t.passthrough_2(e) is e
    assert t.query_shared_1(e) == 5
    assert t.query_shared_2(e) == 5
    assert t.query_shared_3(e) == 5

    w = t.SharedWrapper(e)
    assert e is not w.value
    assert w.value == 5
    w.value = 6
    assert e.value == 6
    del w, e

    e = t.Example.make_shared(6)
    assert t.query_shared_1(e) == 6
    assert t.query_shared_2(e) == 6
    assert t.query_shared_3(e) == 6
    assert t.passthrough(e) is e
    assert t.passthrough_2(e) is e

    w = t.SharedWrapper(e)
    assert e is not w.value
    assert w.value == 6
    del w, e
    collect()
    assert t.stats() == (2, 2)

# ------------------------------------------------------------------

def test04_uniqueptr_from_cpp(clean):
    a = t.unique_from_cpp()
    b = t.unique_from_cpp_2()
    assert a.value == 1
    assert b.value == 2
    del a, b
    collect()
    assert t.stats() == (2, 2)


def test05a_uniqueptr_from_cpp(clean):
    # Test ownership exchange when the object has been created on the C++ side
    a = t.unique_from_cpp()
    b = t.unique_from_cpp_2()
    wa = t.UniqueWrapper(a)
    wb = t.UniqueWrapper(b)
    with pytest.warns(RuntimeWarning, match='nanobind: attempted to access a relinquished instance of type \'test_holders_ext.Example\'!'):
        with pytest.raises(TypeError) as excinfo:
            assert a.value == 1
        assert 'incompatible function arguments' in str(excinfo.value)
    with pytest.warns(RuntimeWarning, match='nanobind: attempted to access a relinquished instance of type \'test_holders_ext.Example\'!'):
        with pytest.raises(TypeError) as excinfo:
            assert b.value == 2
        assert 'incompatible function arguments' in str(excinfo.value)
    del a, b
    del wa, wb
    collect()
    assert t.stats() == (2, 2)

    t.reset()
    a = t.unique_from_cpp()
    b = t.unique_from_cpp_2()
    wa = t.UniqueWrapper(a)
    wb = t.UniqueWrapper(b)

    a2 = wa.get()
    b2 = wb.get()
    assert a2.value == 1 and b2.value == 2
    assert a2 is a and b2 is b
    assert a.value == 1 and b.value == 2
    collect()
    assert t.stats() == (2, 0)
    del a, b, a2, b2
    collect()
    assert t.stats() == (2, 2)

def test05b_uniqueptr_list(clean):
    t.reset()
    # Test ownership exchange as part of a larger data structure
    k = t.unique_from_cpp(1)
    v = t.unique_from_cpp(2)
    res = t.passthrough_unique_pairs([(k, v)])
    assert res == [(k, v)]
    assert k.value == 1 and v.value == 2

    res = t.passthrough_unique_pairs([(k, v)], clear=True)
    assert res == []
    for obj in (k, v):
        with pytest.warns(RuntimeWarning, match='nanobind: attempted to access a relinquished instance of type \'test_holders_ext.Example\'!'):
            with pytest.raises(TypeError) as excinfo:
                obj.value
    collect()
    assert t.stats() == (2, 2)

def test05c_uniqueptr_structure_duplicate(clean):
    t.reset()
    # Test ownership exchange that fails partway through
    # (can't take ownership from k twice)
    k = t.unique_from_cpp(3)
    with pytest.warns(RuntimeWarning, match=r'nanobind::detail::nb_relinquish_ownership()'):
        with pytest.raises(TypeError):
            t.passthrough_unique_pairs([(k, k)])
    # Ownership passes back to Python
    assert k.value == 3

    del k
    collect()
    assert t.stats() == (1, 1)

def test05d_uniqueptr_reinit(clean):
    x = t.unique_from_cpp()
    assert x.value == 1
    w = t.UniqueWrapper(x)
    with pytest.warns(RuntimeWarning, match='nanobind: attempted to access a relinquished instance of type \'test_holders_ext.Example\'!'):
        with pytest.raises(TypeError):
            x.value
    with pytest.warns(RuntimeWarning, match='nanobind: attempted to access a relinquished instance of type \'test_holders_ext.Example\'!'):
        with pytest.raises(TypeError):
            x.__init__(3)
    y = w.get()
    assert y is x and y.value == 1


def test06_uniqueptr_from_py(clean):
    # Test ownership exchange when the object has been created on the Python side
    a = t.Example(1)
    with pytest.warns(RuntimeWarning, match=r'nanobind::detail::nb_relinquish_ownership()'):
        with pytest.raises(TypeError) as excinfo:
            wa = t.UniqueWrapper(a)
    wa = t.UniqueWrapper2(a)
    with pytest.warns(RuntimeWarning, match='nanobind: attempted to access a relinquished instance of type \'test_holders_ext.Example\'!'):
        with pytest.raises(TypeError) as excinfo:
            assert a.value == 1
        assert 'incompatible function arguments' in str(excinfo.value)
    a2 = wa.get()
    assert a2.value == 1 and a is a2
    del a, a2
    collect()
    assert t.stats() == (1, 1)

    # Test that ownership exchange as part of a larger data structure fails
    # gracefully rather than crashing
    k = t.Example(1)
    v = t.Example(2)
    with pytest.warns(RuntimeWarning, match=r'nanobind::detail::nb_relinquish_ownership()'):
        with pytest.raises(TypeError) as excinfo:
            t.passthrough_unique_pairs([(k, v)])
    assert k.value == 1 and v.value == 2

    # Test the case where the key relinquishes ownership successfully and
    # then the value can't do
    v = t.unique_from_cpp(3)
    with pytest.warns(RuntimeWarning, match=r'nanobind::detail::nb_relinquish_ownership()'):
        with pytest.raises(TypeError) as excinfo:
            t.passthrough_unique_pairs([(k, v)])
    assert k.value == 1 and v.value == 3

    del k, v
    collect()
    assert t.stats() == (4, 4)


def test07_uniqueptr_passthrough(clean):
    assert t.passthrough_unique(None) is None
    assert t.passthrough_unique(t.unique_from_cpp()).value == 1
    assert t.passthrough_unique(t.unique_from_cpp_2()).value == 2
    assert t.passthrough_unique_2(t.unique_from_cpp()).value == 1
    assert t.passthrough_unique_2(t.unique_from_cpp_2()).value == 2
    collect()
    assert t.stats() == (4, 4)
    t.reset()

    with pytest.warns(RuntimeWarning, match=r'nanobind::detail::nb_relinquish_ownership()'):
        with pytest.raises(TypeError):
            assert t.passthrough_unique(t.Example(1)).value == 1
    assert t.passthrough_unique_2(t.Example(1)).value == 1
    collect()
    assert t.stats() == (2, 2)


def test07_polymorphic_downcast_unique():
    assert isinstance(t.u_factory(), t.Base)
    assert isinstance(t.u_factory_2(), t.Base)
    assert isinstance(t.u_polymorphic_factory(), t.PolymorphicSubclass)
    assert isinstance(t.u_polymorphic_factory_2(), t.PolymorphicBase)


def test08_polymorphic_downcast_shared():
    assert isinstance(t.s_factory(), t.Base)
    assert isinstance(t.s_factory_2(), t.Base)
    assert isinstance(t.s_polymorphic_factory(), t.PolymorphicSubclass)
    assert isinstance(t.s_polymorphic_factory_2(), t.PolymorphicBase)

def test09_tag_based():
    assert isinstance(t.make_pet(t.PetKind.Dog), t.Dog)
    assert isinstance(t.make_pet(t.PetKind.Cat), t.Cat)


def test09_tag_based_unique():
    assert isinstance(t.make_pet_u(t.PetKind.Dog), t.Dog)
    assert isinstance(t.make_pet_u(t.PetKind.Cat), t.Cat)

def test09_tag_based_shared():
    assert isinstance(t.make_pet_s(t.PetKind.Dog), t.Dog)
    assert isinstance(t.make_pet_s(t.PetKind.Cat), t.Cat)


def check_shared_from_this_py_owned(ty, factory, value):
    e = ty(value)

    # Creating from Python does not enable shared_from_this
    assert e.value == value
    assert not e.has_shared_from_this()
    assert t.owns_cpp(e)

    # Passing to C++ as a shared_ptr does
    w = t.SharedWrapperST(e)
    assert e.has_shared_from_this()
    assert w.ptr is e

    # Execute shared_from_this on the C++ side
    w2 = t.SharedWrapperST.from_existing(e)
    assert e.use_count() == 2
    assert w.value == w2.value == e.value == value
    assert t.same_owner(w, w2)

    # Returning a raw pointer from C++ locates the existing instance
    assert w2.get_own() is w2.get_ref() is e
    assert t.owns_cpp(e)

    if hasattr(sys, "getrefcount"):
        # One reference is held by the C++ shared_ptr, one by our
        # locals dict, and one by the arg to getrefcount
        rc = sys.getrefcount(e)
        assert rc == 3

    # Dropping the Python object does not actually destroy it, because
    # the shared_ptr holds a reference. There is still a PyObject* at
    # the same address.
    prev_id = id(e)
    del e
    collect()
    assert t.stats() == (1, 0)
    assert id(w.get_ref()) == prev_id
    assert t.owns_cpp(w.get_ref())
    assert type(w.get_ref()) is ty

    # Dropping the wrappers' shared_ptrs drops the PyObject reference and
    # the object is finally GC'ed
    del w, w2
    collect()
    assert t.stats() == (1, 1)


def test10_shared_from_this_create_in_python(clean):
    check_shared_from_this_py_owned(t.ExampleST, t.ExampleST, 42)

    # Subclass in C++
    t.reset()
    check_shared_from_this_py_owned(t.DerivedST, t.DerivedST, 30)

    # Subclass in Python
    class SubST(t.ExampleST):
        pass

    t.reset()
    check_shared_from_this_py_owned(SubST, SubST, 20)


def test11_shared_from_this_create_raw_in_cpp(clean):
    # Creating a raw pointer from C++ does not enable shared_from_this;
    # although the object is held by pointer rather than value, the logical
    # ownership transfers to Python and the behavior is equivalent to test10.
    # Once we get a shared_ptr it owns a reference to the Python object.
    check_shared_from_this_py_owned(t.ExampleST, t.ExampleST.make, 10)

    # Subclass in C++
    t.reset()
    check_shared_from_this_py_owned(t.DerivedST, t.DerivedST.make, 5)


def test12_shared_from_this_create_shared_in_cpp(clean):
    # Creating a shared_ptr from C++ enables shared_from_this. Now the
    # shared_ptr does not keep the Python object alive; it's directly
    # owning the ExampleST object on the C++ side.
    e = t.ExampleST.make_shared(10)
    assert e.value == 10
    assert e.has_shared_from_this()
    assert e.shared_from_this() is e  # same instance
    assert e.use_count() == 1
    assert not t.owns_cpp(e)
    if hasattr(sys, "getrefcount"):
        # One reference is held by our locals dict and one by the
        # arg to getrefcount
        rc = sys.getrefcount(e)
        assert rc == 2

    w = t.SharedWrapperST.from_existing(e)
    assert w.ptr is e
    # One shared_ptr whose lifetime is tied to e. And one inside the wrapper
    assert e.use_count() == 2

    # Drop the Python object; C++ object still remains owned by the wrapper
    del e
    collect()
    assert t.stats() == (1, 0)
    assert w.use_count() == 1

    # Get a new Python object reference; it will share ownership of the
    # same underlying C++ object
    e2 = w.get_own()
    assert not t.owns_cpp(e2)
    assert w.ptr is e2
    assert w.use_count() == 2

    del e2
    collect()
    assert t.stats() == (1, 0)
    assert w.use_count() == 1

    # Get a new C++-side reference
    w2 = t.SharedWrapperST.from_wrapper(w)
    assert w2.use_count() == 2
    assert t.same_owner(w, w2)

    # Get another one by roundtripping through Python.
    # The nanobind conversion to shared_ptr<ExampleST> should use the
    # existing shared_from_this shared_ptr
    w3 = t.SharedWrapperST(w.ptr)
    collect()  # on pypy the w.ptr temporary can stay alive
    assert w3.use_count() == 3
    assert t.same_owner(w2, w3)

    # Destroy everything
    assert t.stats() == (1, 0)
    del w, w2, w3
    collect()
    assert t.stats() == (1, 1)


def test13_shared_from_this_create_derived_in_cpp(clean):
    # This tests that keep_shared_from_this_alive is inherited by
    # derived classes properly

    # Pass shared_ptr<T> to Python
    e = t.DerivedST.make_shared(20)
    assert type(e) is t.DerivedST
    assert e.value == 20
    assert e.has_shared_from_this()
    assert not t.owns_cpp(e)
    assert e.use_count() == 1

    # Pass it back to C++
    w = t.SharedWrapperST(e)
    assert e.use_count() == w.use_count() == 2

    del e
    collect()
    assert t.stats() == (1, 0)
    assert w.use_count() == 1

    # Pass it back to Python as a raw pointer
    e = w.get_own()
    # ExampleST is not polymorphic, so the derived-class identity is
    # lost once the Python instance is destroyed
    assert type(e) is t.ExampleST
    assert not t.owns_cpp(e)
    assert w.use_count() == 2
    assert w.get_own() is e

    del e, w
    collect()
    assert t.stats() == (1, 1)


def test14_nullable_properties():
    a = t.ExampleWrapper()
    assert a.value is not None
    assert a.value.value == 5
    assert a.value_nullable is None
    a.value = t.Example(123)
    a.value_nullable = t.Example(456)
    assert a.value.value == 123
    assert a.value_nullable.value == 456
    a.value_nullable = None
    assert a.value_nullable is None
    with pytest.raises(TypeError):
        a.value = None



import sys

if sys.version_info < (3, 11, 0):
    # Work around limitations in typing.overload in Python<3.11
    import typing_extensions as typing
else:
    import typing

# Ignore a type and a function from elsewhere. These shouldn't be included in
# the stub by default
from os import PathLike, getcwd

del sys

C = 123
T = typing.TypeVar("T")

def f1(a, b, c, /):
    """docstring"""

def f2(a=(3,(4,5)), /, b=(1,2), *, c=4):
    """docstring 2"""

def f3(*args, **kwargs):
    pass

def f4() -> typing.Callable[[T], T]:
    return lambda x: x

class AClass:
    STATIC_VAR: int = 5

    class NestedClass:
        pass

    def __init__(self, x):
        pass

    def method(self, x: str):
        pass

    @staticmethod
    def static_method(x):
        pass

    @classmethod
    def class_method(cls, x):
        pass

    @typing.overload
    def overloaded(self, x: int) -> None:
        """docstr 1"""

    @typing.overload
    def overloaded(self, x: str) -> None:
        """docstr 2"""

    def overloaded(self, x):
        pass

    @typing.overload
    def overloaded_2(self, x: int) -> None: ...

    @typing.overload
    def overloaded_2(self, x: str) -> None: ...

    def overloaded_2(self, x):
        "docstr 3"



#!/usr/bin/env python3
import os, re, subprocess, sys

tests = [
    ('demumble hello', 'hello\n'),
    ('demumble _Z4funcPci _Z1fv', 'func(char*, int)\nf()\n'),
    ('demumble < _Z4funcPci _Z1fv', 'func(char*, int)\nf()\n'),
    ('demumble _RINvNtC3std3mem8align_ofdE _RNvNvC5mylib3foo3bar',
     'std::mem::align_of::<f64>\nmylib::foo::bar\n'),
    ('demumble < _RINvNtC3std3mem8align_ofdE _RNvNvC5mylib3foo3bar',
     'std::mem::align_of::<f64>\nmylib::foo::bar\n'),
    ('demumble _TtP3foo3bar_', 'foo.bar\n'),
    ('demumble < _TtP3foo3bar_', 'foo.bar\n'),
    ('demumble $sSS5countSivg', 'Swift.String.count.getter : Swift.Int\n'),
    ('demumble < $sSS5countSivg', 'Swift.String.count.getter : Swift.Int\n'),
    ('demumble ?Fxi@@YAHP6AHH@Z@Z', 'int __cdecl Fxi(int (__cdecl *)(int))\n'),
    ('demumble ??0S@@QEAA@$$QEAU0@@Z', 'public: __cdecl S::S(struct S &&)\n'),
    ('demumble ??_C@_02PCEFGMJL@hi?$AA@', '"hi"\n'),
    ('demumble __Znwi', 'operator new(int)\n'),  # Strip extra _ (for macOS)
    ('demumble < __Znwi', 'operator new(int)\n'),  # Also from stdin
    ('demumble -m hi _Z1fv ho _Z1gv', 'hi\nf()\nho\ng()\n'),
    ('demumble -m < hi_ho _Z1fv ho _Z1gv ?hm', 'f()\ng()\n?hm\n'),
    ('demumble -m < _Z1fv!_Z1gv', 'f()\ng()\n'),
    ('demumble -m < _Z1fv!foo_bar', 'f()\n'),
    ('demumble Pi', 'int*\n'),
    ('demumble < Pi', 'Pi\n'),
    ('demumble < ___Z10blocksNRVOv_block_invoke',
     'invocation function for block in blocksNRVO()\n'),
    ('demumble < .____Z10blocksNRVOv_block_invoke',
     '.invocation function for block in blocksNRVO()\n'),
    ('demumble -m < .____Z10blocksNRVOv_block_invoke',
     'invocation function for block in blocksNRVO()\n'),
    ('demumble _ZN2zx7channelD4Ev', 'zx::channel::~channel()\n'),
    ('demumble -- -b', '-b\n'),
    ('demumble -- -m', '-m\n'),
    ('demumble -- -h', '-h\n'),
    ('demumble -h', re.compile('.*usage: demumble.*')),
    ('demumble --help', re.compile('.*usage: demumble.*')),
    ('demumble --version', re.compile('.*\..*')),
    ('demumble -b hello', 'hello\n'),
    ('demumble -b _Z1fv', '"f()" (_Z1fv)\n'),
    ('demumble -b < _Z1fv', '"f()" (_Z1fv)\n'),
    ('demumble -bm < _Z1fv!foo_bar', '"f()" (_Z1fv)\n'),
    ('demumble -mb < _Z1fv!foo_bar', '"f()" (_Z1fv)\n'),
    ('demumble --foo < bar', re.compile(".*unrecognized option `--foo'.*")),
    ('demumble -bx < bar', re.compile(".*unrecognized option `x' in `-bx'.*")),
    ('demumble < _ZZ3fooiENK3$_0clEi',
     'foo(int)::$_0::operator()(int) const\n'),
    ('demumble .?AVNet@@', "class Net `RTTI Type Descriptor Name'\n"),
    ('demumble < asdf?x@@3HAjkl', 'asdfint xjkl\n'),
    ('demumble < asdf?x@@3Hjkl', 'asdf?x@@3Hjkl\n'),
    ('demumble ?x@@3HAjkl', 'int x\n  unused suffix: jkl\n'),
]

status = 0
for t in tests:
    cmd = t[0].split()
    # Assume that demumble is next to this script.
    cmd[0] = os.path.join(os.path.dirname(__file__) or '.', cmd[0])
    if '<' in cmd:
        p = subprocess.Popen(cmd[:cmd.index('<')], stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                             encoding='utf-8')
        out = p.communicate(input='\n'.join(cmd[cmd.index('<') + 1:]) + '\n')[0]
    else:
        out = subprocess.check_output(cmd, encoding='utf-8')
    if (out != t[1] if isinstance(t[1], str) else not t[1].match(out)):
        print(f"`{t[0]}`: Expected '{t[1]}', got '{out}'")
        status = 1
print('passed' if status == 0 else 'failed')
sys.exit(status)



import test_intrusive_ext as t
import pytest
from common import collect

@pytest.fixture
def clean():
    collect()
    t.reset()

def test01_construct(clean):
    o = t.Test()
    assert o.value(0) == 123
    assert t.get_value_1(o) == 124
    assert t.get_value_2(o) == 125
    assert t.get_value_3(o) == 126
    del o
    collect()
    assert t.stats() == (1, 1)


def test02_factory(clean):
    o = t.Test.create_raw()
    assert o.value(0) == 123
    assert t.get_value_1(o) == 124
    assert t.get_value_2(o) == 125
    assert t.get_value_3(o) == 126
    del o
    collect()
    assert t.stats() == (1, 1)


def test03_factory_ref(clean):
    o = t.Test.create_ref()
    assert o.value(0) == 123
    assert t.get_value_1(o) == 124
    assert t.get_value_2(o) == 125
    assert t.get_value_3(o) == 126
    del o
    collect()
    assert t.stats() == (1, 1)

def test04_subclass(clean):
    class MyTest(t.Test):
        def __init__(self, x):
            super().__init__()
            self.x = x

        def value(self, value):
            return self.x * value

    o = MyTest(456)
    assert o.value(0) == 0
    assert t.get_value_1(o) == 456
    assert t.get_value_2(o) == 456*2
    assert t.get_value_3(o) == 456*3
    del o
    collect()
    assert t.stats() == (1, 1)
